/*! rollup-plugin-userscript v0.3.7 | MIT License */
import { readFile } from 'fs/promises';
import MagicString from 'magic-string';
import { attachScopes } from '@rollup/pluginutils';
import { walk } from 'estree-walker';
import isReference from 'is-reference';

const gmAPIs = ['GM_info', 'GM_getValue', 'GM_getValues', 'GM_setValue', 'GM_setValues', 'GM_deleteValue', 'GM_deleteValues', 'GM_listValues', 'GM_addValueChangeListener', 'GM_removeValueChangeListener', 'GM_getResourceText', 'GM_getResourceURL', 'GM_addElement', 'GM_addStyle', 'GM_openInTab', 'GM_registerMenuCommand', 'GM_unregisterMenuCommand', 'GM_notification', 'GM_setClipboard', 'GM_xmlhttpRequest', 'GM_download'];
const META_START = '// ==UserScript==';
const META_END = '// ==/UserScript==';
function collectGmApi(ast) {
  let scope = attachScopes(ast, 'scope');
  const grantSetPerFile = new Set();
  walk(ast, {
    enter(node, parent) {
      if (node.scope) scope = node.scope;
      if (node.type === 'Identifier' && isReference(node, parent) && !scope.contains(node.name)) {
        if (gmAPIs.includes(node.name)) {
          grantSetPerFile.add(node.name);
        }
      }
    },
    leave(node) {
      if (node.scope) scope = scope.parent;
    }
  });
  return grantSetPerFile;
}
function getMetadata(metaFileContent, additionalGrantList) {
  const lines = metaFileContent.split('\n').map(line => line.trim());
  const start = lines.indexOf(META_START);
  const end = lines.indexOf(META_END);
  if (start < 0 || end < 0) {
    throw new Error('Invalid metadata block. For more details see https://violentmonkey.github.io/api/metadata-block/');
  }
  const grantSet = new Set();
  const entries = lines.slice(start + 1, end).map(line => {
    if (!line.startsWith('// ')) return;
    line = line.slice(3).trim();
    const matches = line.match(/^(\S+)(\s.*)?$/);
    if (!matches) return;
    const key = matches[1];
    const value = (matches[2] || '').trim();
    if (key === '@grant') {
      grantSet.add(value);
      return;
    }
    return [key, value];
  }).filter(Boolean);
  for (const item of additionalGrantList) {
    grantSet.add(item);
  }
  const grantList = Array.from(grantSet);
  grantList.sort();
  for (const item of grantList) {
    entries.push(['@grant', item]);
  }
  const maxKeyWidth = Math.max(...entries.map(([key]) => key.length));
  return [META_START, ...entries.map(([key, value]) => `// ${key.padEnd(maxKeyWidth)} ${value}`), META_END].join('\n');
}

const suffix = '?userscript-metadata';
var index = (transform => {
  const metadataMap = new Map();
  const grantMap = new Map();
  return {
    name: 'userscript-metadata',
    async resolveId(source, importer, options) {
      if (source.endsWith(suffix)) {
        let {
          id
        } = await this.resolve(source, importer, options);
        if (id.endsWith(suffix)) id = id.slice(0, -suffix.length);
        metadataMap.set(importer, id);
        return source;
      }
    },
    load(id) {
      if (id.endsWith(suffix)) {
        return '';
      }
    },
    transform(code, id) {
      const ast = this.parse(code);
      const grantSetPerFile = collectGmApi(ast);
      grantMap.set(id, grantSetPerFile);
    },
    /**
     * Use `renderChunk` instead of `banner` to preserve the metadata after minimization.
     * Note that this plugin must be put after `@rollup/plugin-terser`.
     */
    async renderChunk(code, chunk) {
      const metadataFile = chunk.isEntry && [chunk.facadeModuleId, ...Object.keys(chunk.modules)].map(id => metadataMap.get(id)).find(Boolean);
      if (!metadataFile) return;
      let metadata = await readFile(metadataFile, 'utf8');
      const grantSet = new Set();
      for (const id of this.getModuleIds()) {
        const grantSetPerFile = grantMap.get(id);
        if (grantSetPerFile) {
          for (const item of grantSetPerFile) {
            grantSet.add(item);
          }
        }
      }
      metadata = getMetadata(metadata, grantSet);
      if (transform) metadata = transform(metadata);
      const s = new MagicString(code);
      s.prepend(`${metadata}\n\n`);
      return {
        code: s.toString(),
        map: s.generateMap({
          hires: 'boundary'
        }).toString()
      };
    }
  };
});

export { index as default };
